# Copyright 2020, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of pytest-executable
#     https://www.github.com/CS-SI/pytest-executable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Builtin test module.

This is the module automatically executed when a test_case.yaml file is found.
"""


def test_runner(runner):
    """Check the runner execution.

    An OK process execution shall return the code 0.

    Args:
        runner: Fixture to run the runner script.
    """
    assert runner.run() == 0


def test_logs(output_path):
    """Check the executable log files.

    The error log shall be empty and the output log shall not be empty.

    Args:
        output_path: Path to the current test output directory.
    """
    assert (
        output_path / "executable.stdout"
    ).stat().st_size != 0, "stdout file shall be non-empty"
    assert (
        output_path / "executable.stderr"
    ).stat().st_size == 0, "stderr file shall be empty"



# Copyright 2020, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of pytest-executable
#     https://www.github.com/CS-SI/pytest-executable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides the container for the settings of a test case.

We use a class because a dictionary does not offer easy checking and
code completion.
"""

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Set, Union

from .yaml_helper import YamlHelper

# the yaml file with the default settings is in the same directory as the
# current module, the yaml schema too
SETTINGS_SCHEMA_FILE = Path(__file__).parent / "test_case-schema.yaml"

SettingsDataType = Dict[str, Union[int, List[str], Dict[str, float]]]


@dataclass
class Tolerances:
    """Fields comparison tolerances.

    Attributes:
        rel: The relative tolerance.
        abs: The absolute tolerance.
    """

    rel: float = 0.0
    abs: float = 0.0


@dataclass
class Settings:
    """Test settings.

    This dataclass contains the test settings read from a yaml file.

    Attributes:
        nproc: The number of MPI processes.
        marks: The pytest marks.
        references: The reference files path patterns.
        tolerances: The fields comparison tolerances.
    """

    nproc: int
    marks: Set[str]
    references: Set[str]
    tolerances: Dict[str, Tolerances]

    def __post_init__(self):
        """Coerce the attributes types."""
        self.marks = set(self.marks)
        self.references = set(self.references)
        for key, value in self.tolerances.copy().items():
            self.tolerances[key] = Tolerances(**value)

    @classmethod
    def from_local_file(cls, path_global: Path, path_local: Path) -> "Settings":
        """Create a :class:Settings object from 2 yaml files.

        The contents of the local file overrides or extends the contents of the
        global one.

        Args:
            path_global: Path to a yaml file with global settings.
            path_local: Path to a yaml file with local settings.

        Returns:
            Settings object.
        """
        loader = YamlHelper(SETTINGS_SCHEMA_FILE)
        return cls(**loader.load_merge(path_global, path_local))



# Copyright 2020 CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of pytest-executable
#     https://www.github.com/CS-SI/pytest-executable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide the report database dumper and report generation."""

import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, Set

from _pytest.terminal import TerminalReporter

from .yaml_helper import YamlHelper

YAML_HELPER = YamlHelper(Path(__file__).parent / "report-db-schema.yaml")

REPORT_DB_FILENAME = "tests_report_db.yaml"
REPORT_DIRNAME = "report"

ReportDBType = Dict[str, Dict[str, Any]]


def create(terminalreporter: TerminalReporter) -> ReportDBType:
    """Create the report database.

    A test case is errored if at least one test function is error. A test case
    is failed if at least one test function is failed. A test case is skipped
    if all the test functions are skipped. Otherwise, it is passed.

    All the error and failure messages are listed for a given path.

    Args:
        terminalreporter: Pytest terminal reporter.

    Returns:
        The report database.
    """
    report_db: ReportDBType = defaultdict(dict)
    seen_dirs: Set[str] = set()

    for status in ("error", "failed", "passed", "skipped"):
        stats = terminalreporter.stats.get(status, [])
        for test_report in stats:
            path_from_root = Path(*Path(test_report.fspath).parts[1:])
            dir_path = str(path_from_root.parent)
            db_entry = report_db[dir_path]
            messages = db_entry.setdefault("messages", [])
            if status in ("error", "failed"):
                messages += [test_report.longreprtext]
            if dir_path in seen_dirs:
                continue
            seen_dirs.add(dir_path)
            db_entry["status"] = status

    return dict(report_db)


def merge(db_path: Path, new_db: ReportDBType) -> ReportDBType:
    """Merge a report database into an existing database on disk.

    The entries in db_path but not in new_db are left untouched. The entries in
    both are overwritten with the ones in new_db. The entries only in new_db
    are added.

    Args:
        db_path: Path to a database file.
        new_db: Database to be merged from.

    Returns:
        The merged database.
    """
    merged_db = YAML_HELPER.load(db_path)
    merged_db.update(new_db)
    return merged_db


def dump(report_db_path: Path, terminalreporter: TerminalReporter) -> None:
    """Dump the report database.

    If the database file already exists then new test results are merged with
    the ones in the file.

    The database yaml file have the following format:

    path/to/a/test/case:
      status: the test case status among error, failure, passed and skipped
      messages: list of error and failure messages for all the test functions
                of the test case

    Args:
        report_db_path: Path to the report db file.
        terminalreporter: Pytest terminal reporter.
    """
    report_db = create(terminalreporter)

    if report_db_path.is_file():
        report_db = merge(report_db_path, report_db)

    YAML_HELPER.dump(report_db, report_db_path.open("w"))


def generate(
    script_path: str, output_root: Path, terminalreporter: TerminalReporter
) -> None:
    """Generate the report in the output root.

    The directory that contains script_path is shallow copied in output_root,
    the report is created there.

    Args:
        script_path: Path to the reporter generator script.
        output_root: Path to the test results root directory.
        terminalreporter: Pytest terminal reporter.
    """
    # check that the report generator script is there
    reporter_path = Path(script_path).resolve(True)

    # write the report database
    dump(output_root / REPORT_DB_FILENAME, terminalreporter)

    # generate the report
    subprocess.run(str(reporter_path), cwd=output_root, check=True, shell=True)



# Copyright 2020, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of pytest-executable
#     https://www.github.com/CS-SI/pytest-executable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Entry point into the pytest executable plugin."""

import logging
import sys
from functools import cmp_to_key
from pathlib import Path
from typing import Dict, List, Optional

import _pytest
import py
import pytest
from _pytest.config import Config
from _pytest.terminal import TerminalReporter

from . import report, test_case_yaml
from .file_tools import create_output_directory, find_references, get_mirror_path
from .script_runner import ScriptRunner, get_final_script
from .settings import Settings

LOGGER = logging.getLogger(__name__)

# files to be ignored when creating the output directories symlinks
OUTPUT_IGNORED_FILES = ("__pycache__", "conftest.py", "test_case.yaml")

EXE_RUNNER_NAME = "run_executable"

# file with the test default settings
SETTINGS_PATH = Path(__file__).parent / "test_case.yaml"

# caches the test case directory path to marks to propagate them to all the
# test modules of a test case
_marks_cache: Dict[str, List[str]] = {}


def pytest_addoption(parser):
    """CLI options for the plugin."""
    group = parser.getgroup("executable", "executable testing")
    group.addoption(
        "--runner",
        metavar="PATH",
        help="use the shell script at PATH to run an executable, if omitted then "
        "the executable is not run but the other test processing will be",
    )
    group.addoption(
        "--output-root",
        default="tests-output",
        metavar="PATH",
        help="use PATH as the root directory of the tests output, default: %(default)s",
    )
    group.addoption(
        "--overwrite-output",
        action="store_true",
        help="overwrite existing files in the tests output directories",
    )
    group.addoption(
        "--clean-output",
        action="store_true",
        help="clean the tests output directories before executing the tests",
    )
    group.addoption(
        "--regression-root",
        metavar="PATH",
        help="use PATH as the root directory with the references for the "
        "regression testing, if omitted then the tests using the regression_path "
        "fixture will be skipped",
    )
    group.addoption(
        "--default-settings",
        default=SETTINGS_PATH,
        metavar="PATH",
        help="use PATH as the yaml file with the global default test settings instead "
        "of the built-in ones",
    )
    group.addoption(
        "--report-generator",
        metavar="PATH",
        help="use PATH as the script to generate the test report",
    )

    # change default traceback settings to get only the message without the
    # traceback
    term_rep_options = parser.getgroup("terminal reporting").options
    tb_option = next(
        option for option in term_rep_options if option.names() == ["--tb"]
    )
    tb_option.default = "line"


def pytest_sessionstart(session):
    """Check the cli arguments."""
    getoption = session.config.getoption

    # check options clash
    if getoption("clean_output") and getoption("overwrite_output"):
        msg = "options --clean-output and --overwrite-output are not compatible"
        raise pytest.UsageError(msg)

    # check paths are valid
    for option_name in (
        "runner",
        "default_settings",
        "regression_root",
        "report_generator",
    ):
        path = getoption(option_name)
        try:
            Path(path).resolve(True)
        except FileNotFoundError:
            msg = (
                f"argument --{option_name.replace('_', '-')}: "
                f"no such file or directory: {path}"
            )
            raise pytest.UsageError(msg)
        except TypeError:
            # path is None, i.e. no option is defined
            pass


def _get_parent_path(fspath: py.path.local) -> Path:
    """Return the resolved path to a parent directory.

    Args:
        fspath: Path object from pytest.

    Returns:
        Resolved path to the parent directory of the given pat.
    """
    return Path(fspath).parent.resolve(True)


@pytest.fixture(scope="module")
def create_output_tree(request):
    """Fixture to create and return the path to the output directory tree."""
    getoption = request.config.getoption
    output_root = Path(getoption("output_root"))
    parent_path = _get_parent_path(request.node.fspath)
    output_path = get_mirror_path(parent_path, output_root.resolve())

    try:
        create_output_directory(
            parent_path,
            output_path,
            not getoption("overwrite_output"),
            getoption("clean_output"),
            OUTPUT_IGNORED_FILES,
        )
    except FileExistsError:
        msg = (
            f'the output directory "{output_path}" already exists: either '
            "remove it manually or use the --clean-output option to remove "
            "it or use the --overwrite-output to overwrite it"
        )
        raise FileExistsError(msg)


@pytest.fixture(scope="module")
def output_path(request):
    """Fixture to return the path to the output directory."""
    output_root = Path(request.config.getoption("output_root")).resolve(True)
    return get_mirror_path(_get_parent_path(request.node.fspath), output_root)


def _get_settings(config: _pytest.config.Config, path: Path) -> Settings:
    """Return the settings from global and local test_case.yaml.

    Args:
        config: Config from pytest.
        path: Path to a test case directory.

    Returns:
        The settings from the test case yaml.
    """
    return Settings.from_local_file(
        Path(config.getoption("default_settings")),
        _get_parent_path(path) / SETTINGS_PATH.name,
    )


@pytest.fixture(scope="module")
def tolerances(request):
    """Fixture that provides the settings from global and local test_case.yaml."""
    return _get_settings(request.config, request.node.fspath).tolerances


@pytest.fixture(scope="module")
def runner(request, create_output_tree, output_path):
    """Fixture to run the executable with a runner script.

    This fixture will create a :file:`run_executable.sh` script in the test case
    output directory from the script passed to the pytest command line with the
    option :option:`--runner`. The placeholders {nproc} and {output_path} are
    replaced with their actual values in the written script. The runner object
    created by the fixture can be executed with the :py:meth:`run` method
    which will return the return code of the script execution.

    Returns:
        ScriptRunner object.
    """
    runner_path = request.config.getoption("runner")
    if runner_path is None:
        pytest.skip("no runner provided to --runner")

    # check path
    runner_path = Path(runner_path).resolve(True)

    nproc = _get_settings(request.config, request.node.fspath).nproc

    variables = dict(output_path=output_path, nproc=nproc)
    script = get_final_script(runner_path, variables)
    return ScriptRunner(EXE_RUNNER_NAME, script, output_path)


def _get_regression_path(
    config: _pytest.config.Config, fspath: py.path.local
) -> Optional[Path]:
    """Return the path to the reference directory of a test case.

    None is returned if --regression-root is not passed to the CLI.

    Args:
        config: Config from pytest.
        fspath: Path to a test case directory.

    Returns:
        The path to the reference directory of the test case or None.
    """
    regression_path = config.getoption("regression_root")
    if regression_path is None:
        return None
    return get_mirror_path(
        _get_parent_path(fspath), Path(regression_path).resolve(True)
    )


@pytest.fixture(scope="module")
def regression_path(request):
    """Fixture to return the path of a test case under the references tree."""
    regression_path = _get_regression_path(request.config, request.node.fspath)
    if regression_path is None:
        pytest.skip("no tests references root directory provided to --regression-root")
    return regression_path


def pytest_generate_tests(metafunc):
    """Create the regression_file_path parametrized fixture.

    Used for accessing the references files.

    If --regression-root is not set then no reference files will be provided.
    """
    if "regression_file_path" not in metafunc.fixturenames:
        return

    # result absolute and relative file paths to be provided by the fixture parameter
    # empty means skip the test function that use the fixture
    file_paths = []

    regression_path = _get_regression_path(metafunc.config, metafunc.definition.fspath)

    if regression_path is not None:
        settings_path = metafunc.definition.fspath
        settings = _get_settings(metafunc.config, settings_path)

        if settings.references:
            file_paths = find_references(regression_path, settings.references)

    metafunc.parametrize(
        "regression_file_path",
        file_paths,
        scope="function",
        ids=list(map(str, [f.relative for f in file_paths])),
    )


def pytest_collect_file(parent, path):
    """Collect test cases defined with a yaml file."""
    if path.basename == SETTINGS_PATH.name:
        return TestCaseYamlModule(path, parent)


def pytest_configure(config):
    """Register the possible markers and change default error display.

    Display only the last error line without the traceback.
    """
    config.addinivalue_line(
        "markers", 'slow: marks tests as slow (deselect with -m "not slow")'
    )

    # show only the last line with the error message when displaying a
    # traceback
    if config.getoption("tbstyle") == "auto":
        config.option.tbstyle = "line"


class TestCaseYamlModule(pytest.Module):
    """Collector for tests defined with a yaml file."""

    def _getobj(self):
        """Override the base class method.

        To swap the yaml file with the test_case_yaml.py module.
        """
        # prevent python from using the module cache, otherwise the module
        # object will be the same for all the tests
        del sys.modules[test_case_yaml.__name__]
        # backup the attribute before temporary override of it
        fspath = self.fspath
        self.fspath = py.path.local(test_case_yaml.__file__)
        module = self._importtestmodule()
        # restore the backuped up attribute
        self.fspath = fspath
        # set the test case marks from settings.yaml
        settings = _get_settings(self.config, fspath)
        # store the marks for settings them later
        if settings.marks:
            _marks_cache[fspath.dirname] = settings.marks
        return module


def pytest_exception_interact(node, call, report):
    """Change exception display to only show the test path and error message.

    Avoid displaying the test file path and the Exception type.
    """
    excinfo = call.excinfo
    if excinfo.typename == "CollectError" and str(excinfo.value).startswith(
        "import file mismatch:\n"
    ):
        # handle when a custom test script is used in more than one test case with
        # the same name
        dirname = node.fspath.dirname
        filename = node.fspath.basename
        report.longrepr = (
            f"{dirname}\nshall have a __init__.py because {filename} "
            "exists in other directories"
        )
    else:
        report.longrepr.reprcrash = f"{report.nodeid}: {excinfo.value}"


def pytest_collection_modifyitems(
    session: _pytest.main.Session,
    config: _pytest.config.Config,
    items: List[_pytest.nodes.Item],
) -> None:
    """Change the tests execution order.

    Such that:
    - the tests in parent directories are executed after the tests in children
      directories
    - in a test case directory, the yaml defined tests are executed before the
      others (to handle the output directory creation).
    """
    items.sort(key=cmp_to_key(_sort_yaml_first))
    items.sort(key=cmp_to_key(_sort_parent_last))
    _set_marks(items)


def _sort_yaml_first(item_1: _pytest.nodes.Item, item_2: _pytest.nodes.Item) -> int:
    """Sort yaml item first if in the same directory."""
    path_1 = Path(item_1.fspath)
    path_2 = Path(item_2.fspath)
    if path_1.parent == path_2.parent and (path_1.suffix, path_2.suffix) == (
        ".yaml",
        ".py",
    ):
        return -1
    return 1


def _sort_parent_last(item_1: _pytest.nodes.Item, item_2: _pytest.nodes.Item) -> int:
    """Sort item in parent directory last."""
    dir_1 = Path(item_1.fspath).parent
    dir_2 = Path(item_2.fspath).parent
    if dir_2 in dir_1.parents:
        return -1
    return 1


def _set_marks(items: List[_pytest.nodes.Item]) -> None:
    """Set the marks to the test functions under a test case."""
    for dirname, marks in _marks_cache.items():
        for item in items:
            if item.fspath.dirname != dirname:
                continue
            for mark in marks:
                item.add_marker(mark)


def pytest_terminal_summary(
    terminalreporter: TerminalReporter, exitstatus: int, config: Config
) -> None:
    """Create the custom report.

    In the directory that contains the report generator, the report database is
    created and the report generator is called.
    """
    # path to the report generator
    reporter_path = config.getoption("report_generator")
    if reporter_path is None:
        return

    if not terminalreporter.stats:
        # no test have been run thus no report to create or update
        return

    output_root = Path(config.getoption("output_root"))

    terminalreporter.write_sep("=", "starting report generation")

    try:
        report.generate(reporter_path, output_root, terminalreporter)
    except Exception as e:
        terminalreporter.write_line(str(e), red=True)
        terminalreporter.write_sep("=", "report generation failed", red=True)
    else:
        terminalreporter.write_sep("=", "report generation done")



# Copyright 2020, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of pytest-executable
#     https://www.github.com/CS-SI/pytest-executable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide a yaml file loader."""

from pathlib import Path
from typing import Any, Dict, TextIO

import jsonschema
import yaml

# basic type for the content of a yaml with a mapping at the root level
DataType = Dict[str, Any]


class YamlHelper:
    """Yaml file helper class.

    Can load and validate a yaml file, can also merge tha data 2 yaml files.

    Args:
        schema_path: Path to a schema file used for validating a yaml.
    """

    def __init__(self, schema_path: Path):
        with schema_path.open() as file_:
            self.__schema = yaml.safe_load(file_)

    def load(self, path: Path) -> DataType:
        """Return the validated data from a yaml file.

        Args:
            path: Path to a yaml file.

        Returns:
            The validated data.
        """
        with path.open() as file_:
            data = yaml.safe_load(file_)

        if data is None:
            data = {}

        jsonschema.validate(data, self.__schema)

        return data

    def dump(self, data: DataType, stream: TextIO) -> None:
        """Validate and dump data to a yaml file.

        Args:
            data: Data to be dumped.
            stream: IO stream to be written to.
        """
        jsonschema.validate(data, self.__schema)
        yaml.safe_dump(data, stream)

    def load_merge(self, ref: Path, new: Path) -> DataType:
        """Merge the data of 2 yaml files.

        Args:
            ref: Path to the file to be updated with new.
            new: Path to the file to be merged into ref.

        Return:
            The merged dicitonary.
        """
        return self.__recursive_update(self.load(ref), self.load(new))

    @classmethod
    def __recursive_update(cls, ref: DataType, new: DataType) -> DataType:
        """Merge recursively 2 dictionaries.

        The lists are concatenated.
        """
        for key, value in new.items():
            if isinstance(value, dict):
                ref[key] = cls.__recursive_update(ref.get(key, {}), value)
            elif isinstance(value, list):
                # append the items that were not already in the list
                for val in value:
                    if val not in ref[key]:
                        ref[key] += [val]
            else:
                ref[key] = value
        return ref



# Copyright 2020 CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of pytest-executable
#     https://www.github.com/CS-SI/pytest-executable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Output directory tree creation functions."""

import logging
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import List

LOG = logging.getLogger(__name__)


@dataclass
class FilePath:
    """Relative and absolute file path.

    Attributes:
        relative: Relative path to a file.
        absolute: Absolute path to a file.
    """

    absolute: Path
    relative: Path


def get_mirror_path(path_from: Path, path_to: Path) -> Path:
    """Return the mirror path from a path to another one.

    The mirrored path is determined from the current working directory (cwd)
    and the path_from. The path_from shall be under the cwd. The mirrored
    relative path is the part of path_from that has no common ancestor with
    path_to, without its first parent. For example, given cwd, if path_from is
    c/d/e and path_to is /x/y, then it returns /x/y/d/e (no common ancestor
    between path_from and path_to). If path_to is c/x/y, then cwd/c/x/y/e is
    returned.

    Args:
        path_from: Path to be mirrored from.
        path_to: Root path to be mirrored to.

    Returns:
        Mirror path.

    Raises:
        ValueError: If path_from is not under the cwd.
    """
    cwd = str(Path.cwd().resolve())
    path_from = path_from.resolve(True)

    try:
        relative_from = path_from.relative_to(cwd)
    except ValueError:
        # re-raise with better message
        msg = (
            f"the current working directory {cwd} shall be a parent directory of "
            f"the inputs directory {path_from}"
        )
        raise ValueError(msg)

    # number of directory levels to skip for determining the mirrored relative
    # path
    offset = 1

    try:
        relative_to = path_to.relative_to(cwd)
    except ValueError:
        pass
    else:
        # find the common path part between from and to
        for part_from, part_to in zip(relative_from.parts, relative_to.parts):
            if part_from == part_to:
                offset += 1

    return path_to.joinpath(*relative_from.parts[offset:])


def find_references(ref_dir: Path, ref_files: List[str]) -> List[FilePath]:
    """Return the paths to the references files.

    Args:
        ref_dir: Path to a case directory under the references tree.
        ref_files: Path patterns to the references files.

    Returns:
        Absolute and relative paths from a reference case directory to the
        reference files.
    """
    abs_paths: List[Path] = []
    for ref in ref_files:
        abs_paths += ref_dir.glob(ref)

    if not abs_paths:
        return []

    file_paths: List[FilePath] = []
    for abs_path in abs_paths:
        rel_path = abs_path.relative_to(ref_dir)
        file_paths += [FilePath(abs_path, rel_path)]

    return file_paths


def create_output_directory(
    src_dir: Path,
    dst_dir: Path,
    check_dst: bool,
    clean_dst: bool,
    ignored_files: List[str],
) -> None:
    """Create a directory copy with symbolic links.

    The destination directory is created if it does not exist or if clean_dst
    is true. Only the specified input files will be symlinked, the other files
    will be ignored.

    Args:
        src_dir: Path to the source directory.
        dst_dir: Path to the destination directory.
        check_dst: Whether to check that destination exists.
        clean_dst: Whether to remove an existing destination.
        ignored_files: Files to be ignored when creating the destination
        directory.

    Raises:
        FileExistsError: If the destination directory exists when check_dst is
        true and clean_dst is false.
    """
    # destination checking: force erasing or fail
    if check_dst and dst_dir.is_dir():
        if clean_dst:
            LOG.debug("removing output directory %s", dst_dir)
            shutil.rmtree(dst_dir)
        else:
            raise FileExistsError

    LOG.debug("creating a shallow copy from %s to %s", src_dir, dst_dir)
    _shallow_dir_copy(src_dir, dst_dir, ignored_files)


def _shallow_dir_copy(src_dir: Path, dst_dir: Path, ignored_files: List[str]) -> None:
    """Shallow copy a directory tree.

    Directories are duplicated, files are symlinked. The destination directory
    shall not exists and will be created.

    Args:
        src_dir: Path to the source directory.
        dst_dir: Path to the destination directory.
    """
    dst_dir.mkdir(parents=True, exist_ok=True)
    for src_entry in src_dir.iterdir():
        dst_entry = dst_dir / src_entry.name
        if src_entry.name in ignored_files:
            pass
        elif src_entry.is_dir():
            # directories are not symlinked but created so we can not modify a
            # child file by accident
            _shallow_dir_copy(src_entry, dst_entry, ignored_files)
        else:
            # symlink files, first removing an already existing symlink
            if dst_entry.exists():
                dst_entry.unlink()
            dst_entry.symlink_to(src_entry)



"""Entry point to python packaging.

See https://python-packaging-user-guide.readthedocs.io for more details.
See setup.cfg for other settings.
"""

import setuptools

import versioneer

setuptools.setup(
    version=versioneer.get_version(), cmdclass=versioneer.get_cmdclass(),
)



# Copyright 2020, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of pytest-executable
#     https://www.github.com/CS-SI/pytest-executable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains functions to execute a script in a sub process."""

import logging
import stat
import subprocess
from pathlib import Path
from typing import Any, Dict, Iterator

import jinja2

LOG = logging.getLogger(__name__)


class ScriptExecutionError(Exception):
    """Error for script execution."""


class ScriptRunner:
    """Base class for creating and running scripts.

    Args:
        name: Name of the script.
        script: Content of the script.
        exec_dir: Path to the execution directory.
    """

    # extensions for the script related files
    SCRIPT_EXT = "sh"
    STDOUT_EXT = "stdout"
    STDERR_EXT = "stderr"

    # shell used for executing the shell script
    SHELL = "/usr/bin/env bash"

    def __init__(self, name: str, script: str, exec_dir: Path):
        self._name = name
        self._script = script
        self._exec_dir = exec_dir

    def _get_filenames(self) -> Iterator[str]:
        """Return the script, stdout and stderr filenames."""
        return map(
            lambda s: "{}.{}".format(self._name, s),
            (self.SCRIPT_EXT, self.STDOUT_EXT, self.STDERR_EXT),
        )

    def run(self) -> int:
        """Execute the script.

        The script file is created in the execution directory by dumping the
        contents of the script string.
        The script stdout and stderr are each redirected to files named after
        the script file prefix and suffixed with stdout and stderr.

        Returns:
            The return code of the executed subprocess.

        Raises:
            ScriptExecutionError: If the process fails.
        """
        script_filename, stdout_filename, stderr_filename = self._get_filenames()

        script_path = self._exec_dir / script_filename

        # write the script
        with script_path.open("w") as script_file:
            LOG.debug("writing the shell script %s", script_path)
            script_file.write(self._script)

        # make it executable for the user and the group
        permission = stat.S_IMODE(script_path.stat().st_mode)
        script_path.chmod(permission | stat.S_IXUSR | stat.S_IXGRP)

        # redirect the stdout and stderr to files
        stdout = open(self._exec_dir / stdout_filename, "w")
        stderr = open(self._exec_dir / stderr_filename, "w")

        LOG.debug("executing the shell script %s", script_path)
        cmd = self.SHELL.split() + [script_filename]

        try:
            process = subprocess.run(
                cmd, cwd=self._exec_dir, stdout=stdout, stderr=stderr, check=True
            )
        except subprocess.CalledProcessError:
            # inform about the log files
            raise ScriptExecutionError(
                "execution failure, see the stdout and stderr files in "
                f"{self._exec_dir}"
            )
        finally:
            stdout.close()
            stderr.close()

        return process.returncode


def get_final_script(script_path: Path, variables: Dict[str, Any]) -> str:
    """Return the script with replaced placeholders.

    Args:
        script_path: Path to the script.
        variables: Placeholders to be replaced.

    Returns:
        The final script.

    Raises:
        TypeError if the script cannot be processed.
    """
    try:
        template = jinja2.Template(script_path.read_text())
    except UnicodeDecodeError:
        raise TypeError(f"can't read the script {script_path}") from None
    return template.render(**variables)
